    .module bankswitch

    .globl _bankswitch_get_current_bank
    .globl _flashrom_chip_read_bankswitch
    .globl _flashrom_chip_write_bankswitch
    .globl _flashrom_block_read_bankswitch
    .globl _flashrom_block_write_bankswitch
    .globl _flashrom_block_verify_bankswitch
    .globl _default_mem_bank
    .globl _bank_switch_method
    .globl _una_entry_vector

; RomWBW entry vectors
ROMWBW_OLD_SETBNK  .equ 0xFC06  ; prior to v2.6
ROMWBW_OLD_GETBNK  .equ 0xFC09  ; prior to v2.6
ROMWBW_SETBNK      .equ 0xFFF3  ; v2.6 and later (function vector)
ROMWBW_CURBNK      .equ 0xFFE0  ; v2.6 and later (byte variable)

; UNA BIOS banked memory functions
UNABIOS_ENTRY      .equ 0x08 ; entry vector
UNABIOS_BANKEDMEM  .equ 0xFB ; C register - function number
UNABIOS_BANK_GET   .equ 0x00 ; B register - subfunction number
UNABIOS_BANK_SET   .equ 0x01 ; B register - subfunction number

; P112 Z80182 hardware control registers
P112_DCNTL         .equ 0x32 ; Device control register
P112_BBR           .equ 0x39 ; Bank base register
P112_SCR           .equ 0xEF ; System config register

; N8VEM SBC hardware control registers
N8VEM_MPCL_RAM     .equ 0x78 ; IO address of RAM memory pager configuration latch
N8VEM_MPCL_ROM     .equ 0x7C ; IO address of ROM memory pager configuration latch

    .area _CODE
    .z180

    ; load the page in register HL into the banked region (lower 32K)
loadbank:
    ld a, (_bank_switch_method)
    or a
    jr z, loadbank_romwbw_old
    dec a
    jr z, loadbank_unabios
    dec a
    jr z, loadbank_p112
    dec a
    jr z, loadbank_romwbw_26
    dec a
    jr z, loadbank_n8vem_sbc
    ; well, this is unexpected
    ret
loadbank_n8vem_sbc:
    ld a, l
    out (N8VEM_MPCL_ROM), a
    out (N8VEM_MPCL_RAM), a
    ret
loadbank_romwbw_old:
    ld a, l
    call #ROMWBW_OLD_SETBNK
    ret
loadbank_romwbw_26:
    ld a, l
    call #ROMWBW_SETBNK
    ret
loadbank_unabios:
    ; This is slightly tricky. Normally we'd enter UNA via the RST 8 call,
    ; but we can't do this as we're going to replace the lower part of memory
    ; with flash ROM contents, that does not have the RST 8 vector in place.
    ld bc, #(UNABIOS_BANK_SET << 8 | UNABIOS_BANKEDMEM)
    ex de, hl                   ; move page number into DE
    ld hl, #loadbank_una_return
    push hl                     ; put return address on stack
    ld hl, (_una_entry_vector)
    jp (hl)                     ; simulate call to entry vector
loadbank_una_return:
    ret
loadbank_p112:
    ; map in the EEPROM if HL==0, else unmap it
    ld a, h
    or l
    jr nz, unmap_p112_rom
map_p112_rom:
    di                  ; disable interrupts
    xor a
    out0 (P112_BBR), a  ; map ROM into banked area
    in0 a, (P112_SCR)
    and a, #0xf7        ; enable EEPROM
    out0 (P112_SCR), a
    in0 a, (P112_DCNTL)
    or #0xc0            ; set 3 memory wait states
    out0 (P112_DCNTL), a
    ret
unmap_p112_rom:
    out0 (P112_BBR), l
    in0 a, (P112_SCR)
    or a, #0x08         ; disable EEPROM
    out0 (P112_SCR), a
    out0 (P112_DCNTL), h
    ei                  ; interrupts back on
    ret

    ; return the currently loaded page number
_bankswitch_get_current_bank:
    ld a, (_bank_switch_method)
    or a
    jr z, getbank_romwbw_old
    dec a
    jr z, getbank_unabios
    dec a
    jr z, getbank_p112
    dec a
    jr z, getbank_romwbw_26
    dec a
    jr z, getbank_n8vem_sbc
    ; well, this is unexpected
    ld hl, #0
    ret
getbank_n8vem_sbc:
    ld hl, #0x0080      ; we assume that it's the first page of RAM
    ret
getbank_romwbw_old:
    call #ROMWBW_OLD_GETBNK
    ; returns page number in A
    ld h, #0
    ld l, a
    ret
getbank_romwbw_26:
    ld a, (ROMWBW_CURBNK)
    ld h, #0
    ld l, a
    ret
getbank_unabios:
    ld bc, #(UNABIOS_BANK_GET << 8 | UNABIOS_BANKEDMEM)
    ld hl, #getbank_una_return
    push hl                     ; put return address on stack
    ld hl, (_una_entry_vector)
    jp (hl)                     ; simulate call to entry vector
getbank_una_return:
    ; returns page number in DE
    ex de, hl
    ret
getbank_p112:
    in0 h, (P112_DCNTL)
    in0 l, (P112_BBR)
    ret

selectaddr:
    ; 32-bit address is at sp+4 through sp+7
    ld hl,#6        ; we're interested initially in sp+6 and sp+5
    add hl,sp
    ; compute (address >> 15) & 0x0f
    ld a, (hl)      ; top three bits we want are in here
    and #0x7        ; mask off low three bits
    add a, a        ; shift left one place
    dec hl          ; hl is now sp+5
    bit 7, (hl)     ; test top bit
    jr z, bankready
    or #1           ; set low bit if required
bankready:
    ; now A contains (address >> 15) & 0x0f -- let's select that bank!
    push hl         ; we'll need this in a moment
    ld h, #0
    ld l, a
    call loadbank
    pop hl          ; recover pointer to data on stack
    ld d, (hl)
    res 7, d
    dec hl          ; hl is now sp+4
    ld e, (hl)
    ; now DE contains the offset
    ret

_flashrom_chip_read_bankswitch:
    call selectaddr
    ld a, (de) ; read from the flash in banked memory
    ld l, a
    jr putback

_flashrom_chip_write_bankswitch:
    call selectaddr
    inc hl
    inc hl
    inc hl
    inc hl          ; hl is now sp+6 where the value to write resides
    ld a, (hl)
    ld (de), a      ; write to flash memory 
    jr putback

targetlength:
    ex de, hl       ; banked flash address -> hl
    push ix
    ld ix, #0
    add ix, sp
    ld e, 10(ix)    ; buffer address -> de
    ld d, 11(ix)
    ld c, 12(ix)    ; length -> bc
    ld b, 13(ix)
    pop ix
    ret

_flashrom_block_read_bankswitch:
    call selectaddr
    call targetlength
    ldir            ; copy copy copy
putback:
    push hl
    ; put our memory back in the banked region
    ld hl, (_default_mem_bank)
    call loadbank
    pop hl ; recover value read from flash
    ret

_flashrom_block_verify_bankswitch:
    call selectaddr
    call targetlength
    ; DE = source address in RAM (pointer into buffer)
    ; HL = destination address in flash (pointer into banked memory)
    ; BC = byte counter (# bytes remaining to program)
cmpnext:
    ld a, b
    or c
    jr z, cmpok
    ld a, (de)
    cp (hl)
    jr nz, cmpfail
    inc de
    inc hl
    dec bc
    jr cmpnext
cmpok:
    ld l, #1        ; return true
    jr putback
cmpfail:
    ld l, #0        ; return false
    jr putback


_flashrom_block_write_bankswitch:
    call selectaddr
    call targetlength
writenext:
    ; program a range of bytes in flash;
    ; DE = source address in RAM (pointer into buffer)
    ; HL = destination address in flash (pointer into banked memory)
    ; BC = byte counter (# bytes remaining to program)

    ; put flash into write mode
    ld a, #0xaa
    ld (0x5555), a
    ld a, #0x55
    ld (0x2aaa), a
    ld a, #0xa0
    ld (0x5555), a

    ; program the byte
    ld a, (de)
    ld (hl), a

    ; partial setup for the next byte while the write occurs
nextbyte:
    dec bc
    inc de

    ; check if the next byte is 0xff -- we can skip it (0xff = erased)
    ld a, (de)
    inc a
    jr nz, writewait ; only 0xff + 1 = 0
    ld a, b          ; but are there any bytes left to write?
    or c
    jr z, writewait  ; if this was the last byte, complete normally
    inc hl           ; data sheet does not indicate we must read status from the same address
    jr nextbyte      ; okay, we can safely skip this byte!

    ; wait for programming to complete
writewait:
    ld a, (hl)
    cp (hl)
    jr nz, writewait
    ; data sheet advises double checking with two further reads
    ld a, (hl)
    cp (hl)
    jr nz, writewait

    ; finish setup for the next byte
    inc hl           ; do not do this earlier, to ensure we stay in the ROM address space
    ld a, b
    or c
    jr nz, writenext
    jr putback
